﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Threading.Tasks;

using PT.Scheduler;
using PT.SchedulerDefinitions;
using PT.Scheduler.Simulation.Customizations;
using PT.Transmissions;

namespace PlanetTogether.Sample.SameShiftSetup
{
    internal class CanPauseSetup : PT.Scheduler.Simulation.Customizations.SchedulabilityCustomization
    {
        const string c_udfName = "CanPauseSetup";
        ConcurrentDictionary<BaseId, List<Tuple<DateTime, DateTime>>> m_onlineIntervalsPerResource;

        public override string Name => "CanPauseSetup";
        public override string Description => $"Prevents setup segments from spanning offline intervals while allowing run and post processing sengments to span offline time. None of the segments are allowed to span Cleanout Intervals. This applies to Resources that have a User Defined Field called '{c_udfName}' of type 'Boolean' and value 'True'.";

        protected override void SimulationInitialization(ScenarioDetail a_sd, SchedulabilitySimulationInitializationHelper a_schedulabilityHelper, ScenarioDetail.SimulationType a_simulationType, ScenarioBaseT a_transmission)
        {
            m_onlineIntervalsPerResource = GetOnlineIntervalsPerResource(a_sd);
        }

        protected override bool IsSchedulable(ScenarioDetail a_sd, ScenarioDetail.SimulationType a_simType, long a_clock, long a_simulationClock, BaseResource a_res, BaseActivity a_activity, SchedulableInfo a_schedulableInfo, out long o_postEventTime)
        {
            o_postEventTime = -1;
            if (a_simulationClock >= a_sd.GetPlanningHorizonEnd().Ticks) return true;
            if (!m_onlineIntervalsPerResource.TryGetValue(a_res.Id, out var resOnlineIntervals)) return true;
            if (resOnlineIntervals.Count == 0) return true;
            if (a_schedulableInfo.m_requiredSetupSpan.TimeSpanTicks == 0) return true; //There is no setup.

            DateTime processingStart = new DateTime(a_schedulableInfo.m_scheduledStartDate);
            DateTime setupEnd = new DateTime(a_schedulableInfo.m_setupFinishDate);
            foreach (var onlineInterval in resOnlineIntervals)
            {
                if (onlineInterval.Item1 >= setupEnd) break; // passed relevant intervals
                if (processingStart >= onlineInterval.Item1 && setupEnd <= onlineInterval.Item2)
                {
                    return true; // processing start and finish date fall in the same online interval.
                }
            }

            return false;
        }

        ConcurrentDictionary<BaseId, List<Tuple<DateTime, DateTime>>> GetOnlineIntervalsPerResource(ScenarioDetail a_sd)
        {
            ConcurrentDictionary<BaseId, List<Tuple<DateTime, DateTime>>> resOfflineIntervals = new ConcurrentDictionary<BaseId, List<Tuple<DateTime, DateTime>>>();
            ConcurrentBag<Exception> errs = new ConcurrentBag<Exception>();

            Parallel.ForEach(a_sd.PlantManager.GetResourceList(), (res) => 
            {
                try
                {
                    if (!AppliesToResource(res)) return;
                    resOfflineIntervals[res.Id] = GetContinuousOnlineIntervals(res, a_sd.ClockDate, a_sd.GetPlanningHorizonEnd());
                }
                catch (Exception err)
                {
                    errs.Add(err);
                }
            });
            if (errs.Count > 0)
            {
                throw new AggregateException(errs);
            }
            return resOfflineIntervals;
        }

        public List<Tuple<DateTime, DateTime>> GetContinuousOnlineIntervals(Resource a_res, DateTime a_start, DateTime a_end)
        {
            List<Tuple<DateTime, DateTime>> onlineIntervals = GetIntervalsByType(a_res, a_start, a_end, new HashSet<CapacityIntervalDefs.capacityIntervalTypes>()
            {
                CapacityIntervalDefs.capacityIntervalTypes.NormalOnline,
                CapacityIntervalDefs.capacityIntervalTypes.Overtime,
                CapacityIntervalDefs.capacityIntervalTypes.PotentialOvertime
            });
            MergeIntervals(onlineIntervals);
            List<Tuple<DateTime, DateTime>> offlineIntervals = GetIntervalsByType(a_res, a_start, a_end, new HashSet<CapacityIntervalDefs.capacityIntervalTypes>()
            {
                CapacityIntervalDefs.capacityIntervalTypes.Offline,
                CapacityIntervalDefs.capacityIntervalTypes.Cleanout,
            });
            MergeIntervals(offlineIntervals);

            MergeOfflineIntoOnline(onlineIntervals, offlineIntervals);

            return onlineIntervals;
        }

        /// <summary>
        /// gets a list of start and end dates for the intervals of a specified type.
        /// </summary>
        /// <param name="a_res"></param>
        /// <param name="a_start"></param>
        /// <param name="a_end"></param>
        /// <param name="a_capacityTypes"></param>
        /// <returns></returns>
        static List<Tuple<DateTime, DateTime>> GetIntervalsByType(Resource a_res, DateTime a_start, DateTime a_end, HashSet<CapacityIntervalDefs.capacityIntervalTypes> a_capacityTypes)
        {
            List<Tuple<DateTime, DateTime>> onlineIntervals = new List<Tuple<DateTime, DateTime>>();

            for (int onlineIdx = 0; onlineIdx < a_res.ResourceCapacityIntervals.Count; onlineIdx++)
            {
                ResourceCapacityInterval onlineRci = a_res.ResourceCapacityIntervals[onlineIdx];
                if (onlineRci.StartDateTime > a_end) break;
                if (a_capacityTypes.Contains(onlineRci.IntervalType))
                {
                    if ((onlineRci.StartDateTime <= a_start && onlineRci.EndDateTime > a_start) || (onlineRci.StartDateTime >= a_start && onlineRci.StartDateTime < a_end))
                    {
                        DateTime start = onlineRci.StartDateTime <= a_start ? a_start : onlineRci.StartDateTime;
                        DateTime end = onlineRci.EndDateTime > a_end ? a_end : onlineRci.EndDateTime;
                        onlineIntervals.Add(new Tuple<DateTime, DateTime>(start, end));
                    }
                }
            }

            return onlineIntervals;
        }

        /// <summary>
        /// Combines overlapping intervals.
        /// </summary>
        /// <param name="a_intervals"></param>
        static void MergeIntervals(List<Tuple<DateTime, DateTime>> a_intervals)
        {
            int i = 0;
            while (i < a_intervals.Count - 1)
            {
                var interval = a_intervals[i];
                var nextInterval = a_intervals[i + 1];
                if (nextInterval.Item1 <= interval.Item2) // next interval starts before or on the end of this one.
                {
                    if (nextInterval.Item2 > interval.Item2) // next interval extends this one
                    {
                        a_intervals[i] = new Tuple<DateTime, DateTime>(interval.Item1, nextInterval.Item2);
                    }
                    a_intervals.RemoveAt(i + 1); // even if next interval doesn't extend, then it's contained so remove it anyway.
                }
                else // nextInterval.Start > interval.End, move on
                {
                    i++;
                }
            }
        }

        /// <summary>
        /// Adjusts online intervals shortening them when an offline interval overlaps with them or breaking them into two when an offline interval is contained in them or
        /// removing them when they're contained by an offline interval.
        /// </summary>
        /// <param name="a_onlineIntervals"></param>
        /// <param name="a_offlineIntervals"></param>
        static void MergeOfflineIntoOnline(List<Tuple<DateTime, DateTime>> a_onlineIntervals, List<Tuple<DateTime, DateTime>> a_offlineIntervals)
        {
            int i = 0;
            while (i < a_onlineIntervals.Count)
            {
                var online = a_onlineIntervals[i];
                foreach (var offline in a_offlineIntervals)
                {
                    if (offline.Item1 >= online.Item2) break; // this and remaining offlines are past this online.
                    if (offline.Item1 <= online.Item1 && offline.Item2 > online.Item1) // offline starts before online and it ends after online start
                    {
                        if (offline.Item2 > online.Item2) // offline ends after online end (covers the whole thing)
                        {
                            a_onlineIntervals.RemoveAt(i);
                            i--; // subtract since the next interval now has the same index as current.
                            break;
                        }
                        else // offline overlaps the beginning of online
                        {
                            a_onlineIntervals[i] = new Tuple<DateTime, DateTime>(offline.Item2, online.Item2);
                        }
                    }
                    else if (offline.Item1 >= online.Item1 && offline.Item1 < online.Item2) // offline starts somewhere in online interval
                    {
                        if (offline.Item2 < online.Item2) // offline ends before online end (offline is contained in online)
                        {
                            a_onlineIntervals[i] = new Tuple<DateTime, DateTime>(online.Item1, offline.Item1);
                            a_onlineIntervals.Insert(i + 1, new Tuple<DateTime, DateTime>(offline.Item2, online.Item2));
                            break;
                        }
                        else // offline overlaps the end of online
                        {
                            a_onlineIntervals[i] = new Tuple<DateTime, DateTime>(online.Item1, offline.Item1);
                        }
                    }
                }

                i++;
            }
        }

        static bool AppliesToResource(Resource a_res)
        {
            try
            {
                if (a_res.UserFields == null || !a_res.UserFields.Contains(c_udfName)) return false;
                return (bool)a_res.UserFields.Find(c_udfName).DataValue;
            }
            catch
            {
                return false;
            }
        }
    }
}
